# SPDX-FileCopyrightText: 2022 Thomas Kramer
#
# SPDX-License-Identifier: GPL-3.0-or-later

from .parser import parse_liberty
from .types import *


def test_select_timing_group_1():
    data = r"""
pin(Y){ 
    timing() {
        test_label: 1;
        related_pin: "A";
        when: "A & B";
        cell_rise() {
            test_label: 11;
        }
    }
    timing() {
        test_label: 2;
        related_pin: "A";
        when: "!B";
        cell_rise() {
            test_label: 21;
        }
    }
    timing() {
        test_label: 3;
        related_pin: B; // Unescaped string
        when: A; // Unescaped string
        cell_rise() {
            test_label: 31;
        }
    }
}
"""
    pin_group = parse_liberty(data)
    assert isinstance(pin_group, Group)

    timing_group = select_timing_group(pin_group, related_pin="A")
    assert timing_group['test_label'] == 1

    timing_group = select_timing_group(pin_group, related_pin="A", when='B & A') # Note: A and B are swapped in `when` expression.
    assert timing_group['test_label'] == 1

    timing_group = select_timing_group(pin_group, related_pin="A", when='!B')
    assert timing_group['test_label'] == 2

    timing_group = select_timing_group(pin_group, related_pin="B")
    assert timing_group['test_label'] == 3

    assert select_timing_table(pin_group, related_pin="A", when='!B', table_name='cell_rise')['test_label'] == 21
    # Test with unescaped strings.
    assert select_timing_table(pin_group, related_pin="B", when='A', table_name='cell_rise')['test_label'] == 31

def test_replace_array():
    """
    'set_array' should replace existing arrays instead of appending a new one.
    
    See: https://codeberg.org/tok/liberty-parser/issues/16
    """

    data = r"""
    group() {
        myArray ( \
    "0, 0, 0", \
    "0, 0, 0" \
    );
    }
    """

    group = parse_liberty(data)
    assert isinstance(group, Group)

    assert len(group.attributes) == 1

    group.set_array("myArray", np.array([1, 2, 3]))

    assert len(group.attributes) == 1
    assert (group.get_array("myArray") == [1, 2, 3]).all()


def test_select_timing_group_by_timing_type():
    """
    Select timing groups by their `timing_type` attribute.

    Test fix proposed in https://codeberg.org/tok/liberty-parser/issues/16.
    """

    data = r"""
        pin(Y) {
            timing() {
                related_pin : "CLK";
                timing_type : hold_falling;
            }
            timing() {
                related_pin : "CLK";
                timing_type : setup_falling;
            }
            timing() {
                related_pin : "CLK";
                timing_type : "fantasy_escaped_string";
            }
        }    
    """

    pin = parse_liberty(data)
    assert isinstance(pin, Group)


    timing_hold_falling = select_timing_group(pin, related_pin = "CLK", timing_type = "hold_falling")
    assert isinstance(timing_hold_falling, Group)
    assert timing_hold_falling.get_attribute("timing_type") == "hold_falling"

    timing_setup_falling = select_timing_group(pin, related_pin = "CLK", timing_type = "setup_falling")
    assert isinstance(timing_setup_falling, Group)
    assert timing_setup_falling.get_attribute("timing_type") == "setup_falling"

    # Test if timing type is escaped.
    should_not_fail = select_timing_group(pin, related_pin = "CLK", timing_type = "fantasy_escaped_string")
    
    try:
        should_fail = select_timing_group(pin, related_pin = "CLK", timing_type = "doesnotexist")
    except Exception as e:
        # Error message should hint to existing timing types.
        assert "hold_falling" in str(e)
    



def test_library_colon_in_group_argument():
    """
    See https://codeberg.org/tok/liberty-parser/issues/15
    """

    data = r"""
        library(mylib) {
            input_ccb (FOO:a) {}
        }
    """

    lib = parse_liberty(data)
    assert isinstance(lib, Group)


def test_library_name_begins_with_digit():
    """
    See https://codeberg.org/tok/liberty-parser/issues/17
    """

    data = r"""
        library(0V95XXX) {
        
        }
    """

    lib = parse_liberty(data)
    assert isinstance(lib, Group)
    
def test_library_name_with_minus():
    """
    See issue 18.
    """

    data = r"""
        library(some_lib_-10C) {
        
        }
    """

    lib = parse_liberty(data)
    assert isinstance(lib, Group)

def test_format_multiline_string():
    """
    See https://codeberg.org/tok/liberty-parser/issues/19
    """

    data = r"""somegroup () {
  table : "line 1, \
line 2, \
line 3";
}"""
    group = parse_liberty(data)
    assert isinstance(group, Group)

    expected = r"""line 1, \
line 2, \
line 3"""
    assert group["table"] == expected

    # Format again and check for equality with original input.
    formatted = str(group)

    assert formatted == data


def test_without_space_after_colon():
    """
    See https://codeberg.org/tok/liberty-parser/issues/21
    """

    data = r"""
    timing(){ 
        timing_type :"min_pulse_width"; 
    }
    """
    
    group = parse_liberty(data)
    assert isinstance(group, Group)

#def test_two_dimensional_bus_pins():
#    """
#    See: https://codeberg.org/tok/liberty-parser/issues/22
#    """
#
#    data = r"""
#    library(test) {
#        cell(somecell) {
#          bus( x_if[0].y ) {
#            pin("x_if[0].y[0]") {
#                content: asdf ;
#            }
#          }
#        }
#    }
#    """
#
#    group = parse_liberty(data)
#    assert isinstance(group, Group)

def test_semicolon_after_group():
    """
    See https://codeberg.org/tok/liberty-parser/issues/24
    """

    data = r"""
 group( ) {
     inner1() {};
     inner2() {}
 };
    """
    
    group = parse_liberty(data)
    assert isinstance(group, Group)

def test_empty_escaped_string():
    """
    See https://codeberg.org/tok/liberty-parser/issues/25
    """

    data = r"""
    group() {
        comment: "";
    }
"""
    group = parse_liberty(data)
    assert isinstance(group, Group)
    assert group["comment"] == EscapedString("")

def test_issue50_pin_name_with_slash():
    # See: https://codeberg.org/tok/liberty-parser/issues/50

    data ="""library (mylib){
    cell (mycell){
        pinname : a/b ;
        commentorname1a : /*comment*/ name ;
        commentorname1b : /*comment*/name ;
        commentorname2a : name /*comment*/ ;
        commentorname2b : name/*comment*/ ; // This is not a comment.
    }
}
    """

    group = parse_liberty(data)
    assert isinstance(group, Group)
    assert group.get_group("cell", "mycell")["pinname"] == "a/b"
    assert group.get_group("cell", "mycell")["commentorname1a"] == "name"
    assert group.get_group("cell", "mycell")["commentorname1b"] == "name"
    assert group.get_group("cell", "mycell")["commentorname2a"] == "name"
    assert group.get_group("cell", "mycell")["commentorname2b"] == "name/*comment*/"



# SPDX-FileCopyrightText: 2024 Thomas Kramer
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Iterable, Callable

class PeekingIterator:

    def __init__(self, iter: Iterable):
        assert isinstance(iter, Iterable)
        self.iter = iter
        self.has_peeked_value: bool = False
        self.peeked_value = None

    def __iter__(self):
        return self
    
    def __next__(self):
        result = None
        if self.has_peeked_value:
            result = self.peeked_value
            self.peeked_value = None
            self.has_peeked_value = False
        else:
            result = next(self.iter)

        return result

    def peek(self):
        if not self.has_peeked_value:
            self.peeked_value = next(self.iter)
            self.has_peeked_value = True
        return self.peeked_value

        
def test_peekable_iter():
    l = [1, 2, 3]
    i = PeekingIterator(iter(l))

    assert i.peek() == 1
    assert i.peek() == 1
    assert next(i) == 1
    assert i.peek() == 2
    assert next(i) == 2
    assert next(i) == 3
    
class LibertyLexer:

    def __init__(self):
        self.default_terminal_chars = bytearray(b',{}()[];:')
        self.set_default_terminal_chars()

    def set_default_terminal_chars(self):
        self.terminal_chars = self.default_terminal_chars.copy()
        
    def enable_terminal_char(self, char):
        assert len(char) == 1
        c = ord(char)
        if c not in self.terminal_chars:
            self.terminal_chars.append(c)

    def disable_terminal_char(self, char):
        self.terminal_chars.remove(ord(char))
    
    def consume_next_token(self, iter: PeekingIterator, output_fn: Callable):
        try:
            while True:
                self._skip_whitespace(iter)

                c = next(iter)

                if c == '#':
                    # Skip comment
                    self._consume_line(iter)
                elif c == '/':
                    if iter.peek() == '*':
                        next(iter)
                        self._skip_comment_multiline(iter)
                    elif iter.peek() == '/':
                        next(iter)
                        self._consume_line(iter)
                    else:
                        output_fn(c)
                elif c == '\\':
                    # Escape character
                    if iter.peek() == '\\':
                        output_fn('\\')
                        next(iter)
                    elif iter.peek() == '\n':
                        # Ignore masked newline.
                        next(iter)
                    elif iter.peek() == '\r':
                        # Ignore masked newline.
                        next(iter)
                        if iter.peek() == '\n':
                            next(iter)
                    else:
                        output_fn('\\')
                elif c in ['"', "'"]:
                    # Quoted strings.
                    quote_char = c
                    return self._read_quoted_string(quote_char, output_fn, iter)
                else:
                    # Normal token
                    first_char = c
                    return self._read_normal_token(first_char, output_fn, iter)
                        
        except StopIteration:
            pass
            

    def _skip_whitespace(self, iter):
        try:
            while self._is_whitespace(iter.peek()):
                next(iter)
        except StopIteration:
            pass

    def _skip_comment_multiline(self, iter):
        while True:
            # Consume until next '*'
            for c in iter:
                if c == '*':
                    break

            try:
                if iter.peek() == '/':
                    next(iter)
                    # End of comment
                    break
            except StopIteration:
                # End of file.
                break

    def _consume_line(self, iter):
        try:
            for c in iter:
                if c == '\n':
                    break
        except StopIteration:
            pass

    def _is_whitespace(self, c):
        return c.isspace()
    
    def _is_terminal_char(self, c):
        c = ord(c)
        return c < 256 and c in self.terminal_chars

    def _read_quoted_string(self, quote_char, output_fn, iter):
        output_fn(quote_char)
      
        prev = None
        while True:
            c = next(iter)
            output_fn(c)
            if prev != '\\' and c == quote_char:
                # Abort on umasked quote char.
                break
            prev = c
    
    def _read_normal_token(self, first_char, output_fn, iter):
        output_fn(first_char)
        prev = first_char
        if self._is_terminal_char(first_char):
            return

        try:
            while True:
                if self._is_terminal_char(iter.peek()):
                    break
                c = next(iter)
                if prev != '\\' and (self._is_whitespace(c) or self._is_terminal_char(c)):
                    # Abort on unmasked whitespace or terminal character.
                    break
                output_fn(c)
                prev = c
                
        except StopIteration:
            pass
                

def test_skip_comments():

    data = """
      # single line comment 1

      // single line comment 2

     /* multi
    line
    comment */
    """

    lex = LibertyLexer()

    it = PeekingIterator(iter(data))

    buf = []

    def output_fn(c):
        buf.append(c)

    lex.consume_next_token(it, output_fn)
    assert buf == []
    
def test_read_tokens():

    data = r"""
      # single line comment 1

     library (library_name){
        var1 : "quoted_string1";
        var2 : 'quoted_string2';
      // single line comment 2

    /**/
    
    # Masked newline should be ignored.
    \
    
     /* multi
    line
    comment */

    }
    """

    lex = LibertyLexer()

    it = PeekingIterator(iter(data))

    tokens = []
    while True:
        
        buf = []

        def output_fn(c):
            buf.append(c)

        lex.consume_next_token(it, output_fn)

        if not buf:
            break # End of file.
 
        tokens.append("".join(buf)) # Convert to string

    expected_tokens = [
        "library", "(", "library_name", ")", "{",
         "var1", ":", '"quoted_string1"', ";",
         "var2", ":", "'quoted_string2'", ";",
         "}"
    ]

    assert tokens == expected_tokens



# Copyright (c) 2019-2021 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer
#
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Any, List, Dict, Optional, Tuple
from itertools import chain
from .boolean_functions import parse_boolean_function, format_boolean_function
from .arrays import strings_to_array, array_to_strings
import numpy as np
from sympy.logic import boolalg
import sympy


class Define:
    def __init__(self, attribute_name, group_name, attribute_type):
        """

        :param attribute_name: Name of the new defined attribute.
        :param group_name: Name of the group in which the attribute is created.
        :param attribute_type: Data type of the attribute: boolean, string, integer or float
        """

        self.attribute_name = attribute_name
        self.group_name = group_name
        self.attribute_type = attribute_type

    def __str__(self):
        return 'define ("{}", "{}", "{}")'.format(self.attribute_name, self.group_name, self.attribute_type)

    def __repr__(self):
        return str(self)


class Attribute:
    def __init__(self, attribute_name: str, value):
        """
        :param attribute_name: Name of the attribute.
        :param value: Value of the attribute.
        An attribute value can have types
        `float`, `string`, `EscapedString`, `ArithExpression`, `WithUnit`, `NameBitSelection`, `List[float]`.
        """

        self.name = attribute_name
        self.value = value

    def __str__(self):
        return '{} : {}'.format(self.name, self.value)

    def __repr__(self):
        return str(self)


class Group:
    def __init__(self, group_name: str,
                 args: List[str] = None,
                 attributes: List[Attribute] = None,
                 groups: List = None,
                 defines: List[Define] = None):
        self.groups: List[Group] = groups if groups is not None else []

        self.group_name = group_name
        self.args = args if args is not None else []
        assert isinstance(self.args, list)

        self.attributes = attributes if attributes is not None else list()
        assert isinstance(self.attributes, list)
        assert all((isinstance(v, Attribute) for v in self.attributes))

        self.groups = groups if groups is not None else []
        assert isinstance(self.groups, list)
        assert all((isinstance(v, Group) for v in self.groups))

        self.defines: List[Define] = defines if defines is not None else []
        assert isinstance(self.defines, list)
        assert all((isinstance(v, Define) for v in self.defines))

    def get_groups(self, type_name: str, argument: Optional[str] = None) -> List:
        """ Get all groups of type `type_name`.
        Optionally filter the groups by their first argument.
        :param type_name:
        :param argument:
        :return: List[Group]
        """
        return [g for g in self.groups
                if g.group_name == type_name
                and (argument is None or
                     (len(g.args) > 0 and g.args[0] == argument)
                     )
                ]

    def get_group(self, type_name: str, argument: Optional[str] = None):
        """
        Get exactly one group of type `type_name`.
        :param type_name:
        :return: Group
        """
        groups = self.get_groups(type_name, argument=argument)
        assert len(groups) == 1, "There must be exactly one instance of group '{}'. " \
                                 "Found {}.".format(type_name, len(groups))
        return groups[0]

    def __repr__(self) -> str:
        return "%s (%s) t{%s, %s}" % (self.group_name, self.args, self.attributes, self.groups)

    def __str__(self) -> str:
        """
        Create formatted string representation that can be dumped to a liberty file.
        :return:
        """
        return "\n".join(self._format())

    def _format(self, indent: str = " " * 2) -> List[str]:
        """
        Create the liberty file format line by line.
        :return: A list of lines.
        """

        def format_value(v) -> str:
            if v is None:
                return ''
            return str(v)

        define_lines = list()
        for d in self.defines:
            define_lines.append('{};'.format(d))

        sub_group_lines = [g._format(indent=indent) for g in self.groups]

        attr_lines = list()

        for attr in self.attributes:
            attr_name, attr_value = attr.name, attr.value
            if isinstance(attr_value, list):
                # Complex attribute
                formatted = [format_value(x) for x in attr_value]

                if any((isinstance(x, EscapedString) for x in attr_value)):
                    attr_lines.append('{} ( \\'.format(attr_name))
                    for i, l in enumerate(formatted):
                        if i < len(formatted) - 1:
                            end = ', \\'
                        else:
                            end = ' \\'
                        attr_lines.append(indent + l + end)
                    attr_lines.append(');')
                else:
                    values = "({})".format(", ".join(formatted))
                    attr_lines.append("{} {};".format(attr_name, values))
            else:
                # Simple attribute
                values = format_value(attr_value)
                attr_lines.append("{} : {};".format(attr_name, values))

        lines = list()
        lines.append("{} ({}) {{".format(self.group_name, ", ".join([format_value(f) for f in self.args])))

        for l in chain(define_lines, attr_lines, *sub_group_lines):
            lines.append(indent + l)

        lines.append("}")

        return lines

    def get_attributes(self, key: str) -> List[Any]:
        """
        Find attributes values by attribute name.
        :param key: The name of the attribute.
        :return: Returns a list of attribute values.
        """
        return [a.value for a in self.attributes if a.name == key]

    def get_attribute(self, key: str, default=None) -> Any:
        """
        Find exactly one attribute value based on its name.
        Raises an exception if there is no or more than one attributes with this name.
        :param key: Name of the attribute.
        :param default: Returns this default value if no attribute with this name is found.
        :return: The attribute value.
        """
        attrs = self.get_attributes(key)

        if len(attrs) == 0:
            return default

        assert len(attrs) == 1, "Expected to find exactly one attribute with name '{}'. " \
                                "Found {}.".format(key, len(attrs))
        return attrs[0]

    def __getitem__(self, item):
        return self.get_attribute(item)

    def __setitem__(self, key, value):
        """
        Replace the first occurrence of the attribute with the given value.
        """
        
        for a in self.attributes:
            if a.name == key:
                a.value = value
                return
            
        # Otherwise append a new attribute.
        self.attributes.append(Attribute(key, value))

    def __delitem__(self, key):
        """
        Delete all attributes with the name `key`.
        """
        self.attributes = [attr for attr in self.attributes if attr.name != key]

    def __contains__(self, item):
        return len(self.get_attributes(item)) > 0

    def get(self, key, default=None):
        return self.get_attribute(key, default)

    def get_array(self, key) -> np.ndarray:
        """
        Helper function for reading array falues.
        Get a 1D or 2D array as a numpy.ndarray object.
        :param key: Name of the attribute.
        :return: ndarray
        """
        str_array = self[key]
        str_array = [s.value for s in str_array]
        return strings_to_array(str_array)

    def set_array(self, key, value: np.ndarray):
        """
        Helper functions for storing array values.
        Set the value of the attribute "key" to the given array value.
        If attributes with the given name already exist, then the first of them will be replaced.
        """
        str_array = array_to_strings(value)
        str_array = [EscapedString(s) for s in str_array]
        self[key] = str_array

    def get_boolean_function(self, key) -> boolalg.Boolean:
        """
        Get parsed boolean expression.
        Intended for getting the value of the `function` attribute of pins.
        :param key:
        :return: Returns none if there is no function defined under this key.
        """

        f_str = self.get(key)
        if f_str is None:
            return None
        f = parse_boolean_function(f_str.value)
        return f

    def set_boolean_function(self, key, boolean: boolalg.Boolean):
        """
        Format the boolean expression and store it as an attribute with name `key`.
        :param key:
        :param boolean: Sympy boolean expression.
        """
        f_str = format_boolean_function(boolean)
        self[key] = '"{}"'.format(f_str)


class CellGroup(Group):

    def __init__(self, cell_name: str, attributes: List[Attribute],
                 sub_groups: List[Group]):
        super().__init__("cell", args=[cell_name], attributes=attributes,
                         groups=sub_groups)
        self.name = cell_name


class WithUnit:
    """
    Store a value with a unit attached.
    """

    def __init__(self, value, unit: str):
        self.value = value
        self.unit = unit

    def __str__(self):
        return "{}{}".format(self.value, self.unit)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if not isinstance(other, WithUnit):
            return False
        else:
            return self.value == other.value and self.unit == other.unit


class EscapedString:

    def __init__(self, value):
        self.value = value

    def __str__(self):
        return '"{}"'.format(self.value)

    def __repr__(self):
        return str(self)

    def __eq__(self, other):
        if isinstance(other, EscapedString):
            return self.value == other.value
        else:
            return self.value == other

    def __hash__(self):
        return hash(self.value)


class ArithExpression:
    """
    Arithmetic expression like `(VDD + 0.5) * 1.2`.
    Arithmetic expressions are stored as strings.
    """

    def __init__(self, value: str):
        self.value = value

    def __str__(self):
        return '"{}"'.format(self.value)

    def __repr__(self):
        return str(self)

    def to_sympy_expression(self) -> sympy.Expr:
        return sympy.parse_expr(self.value)


class NameBitSelection:
    """Name with bit selection (e.g. ADR[32:0])"""

    def __init__(self, name, sel1, sel2=None):
        self.name = name
        self.sel1 = sel1
        self.sel2 = sel2

    def __str__(self):
        if self.sel2 is not None:
            sel = "{}:{}".format(int(self.sel1), int(self.sel2))
        else:
            sel = int(self.sel1)
        return "{}[{}]".format(self.name, sel)

    def __repr__(self):
        return str(self)


def select_cell(library: Group, cell_name: str) -> Optional[Group]:
    """
    Select a cell by name from a library group.
    :param library:
    :param cell_name:
    :return:
    """
    available_cell_names = {g.args[0] for g in library.get_groups('cell')}

    if cell_name in available_cell_names:
        return library.get_group('cell', cell_name)
    else:
        raise KeyError("Cell name must be one of: {}".format(list(sorted(available_cell_names))))


def select_pin(cell: Group, pin_name: str) -> Optional[Group]:
    """
    Select a pin by name from a cell group.
    :param cell:
    :param pin_name:
    :return:
    """
    available_pin_names = {g.args[0] for g in cell.get_groups('pin')}

    if pin_name in available_pin_names:
        return cell.get_group('pin', pin_name)
    else:
        raise KeyError("Pin name must be one of: {}".format(list(sorted(available_pin_names))))


def select_timing_group(pin: Group,
                        related_pin: str,
                        when: Optional[str] = None,
                        timing_type: Optional[str] = None) -> Optional[Group]:
    """
    Select a timing group by name from a pin group.
    :param pin:
    :param related_pin:
    :param when:
    :param timing_type: Select by 'timing_type' attribute.
    :return:
    """

    assert isinstance(pin, Group)
    assert pin.group_name == 'pin', "`pin` must be a pin-group but is '{}'".format(pin.group_name)

    if when is not None:
        when = parse_boolean_function(when)
        
    def unwrap_escaped_string(s):
        if isinstance(s, EscapedString):
            return s.value
        return s
    
    # Select timing groups by value of the 'related_pin' attribute.
    timing_groups = [g
                     for g in pin.get_groups('timing')
                     if 'related_pin' in g
                     and unwrap_escaped_string(g['related_pin']) == related_pin
                     ]

    if not timing_groups:
        # Notify the user what `related_pin`s could have been chosen.
        raise KeyError(("No timing group found. Related pin name must be one of: {}".
            format(sorted(list({
            g['related_pin'].value
            for g in pin.get_groups('timing')
            if 'related_pin' in g
        })))
        ))

    # Select by `when`.
    if when is not None:

        def expr_equal(a, reference) -> str:
            """
            Test if the two expressions are equal.
            """
            a = unwrap_escaped_string(a)
            a = parse_boolean_function(a)
            return a == reference
            
        timing_groups_selected = [g
                         for g in timing_groups
                         if 'when' in g
                         and expr_equal(g['when'], when)
                         ]
        if not timing_groups_selected:
            # Notify the user what `related_pin`s could have been chosen.
            raise KeyError(("No timing group found. `when` must be one of: {}".
                format(sorted(list({
                unwrap_escaped_string(g['when'])
                for g in timing_groups
                if 'when' in g
            })))
            ))
        timing_groups = timing_groups_selected

    # Select by timing_type
    if timing_type is not None:
        timing_groups_selected = [g
                         for g in timing_groups
                         if 'timing_type' in g
                         and g['timing_type'] == timing_type
                         ]
        if not timing_groups_selected:
            # Notify the user what `timing_type`s could have been chosen.
            raise KeyError(("No timing group found. `timing_type` must be one of: {}".
                format(sorted(list({
                unwrap_escaped_string(g['timing_type'])
                for g in timing_groups
            })))
            ))
        timing_groups = timing_groups_selected

    timing_group = timing_groups[0]
    return timing_group


def select_timing_table(pin: Group,
                        related_pin: str,
                        table_name: str,
                        when: Optional[str] = None,
                        timing_type: Optional[str] = None) -> Optional[Group]:
    """
    Get a timing table by name from a pin group.
    :param pin: The pin group of the output pin.
    :param related_pin: The related input pin name.
    :param table_name: The name of the timing table. ('cell_rise', 'cell_fall', 'rise_transition', ...)
    :param timing_type: Select by 'timing_type' attribute.
    :return:
    """

    # Find timing group.
    timing_group = select_timing_group(pin, related_pin, timing_type=timing_type, when=when)

    available_table_names = {g.group_name for g in timing_group.groups}

    if table_name in available_table_names:
        return timing_group.get_group(table_name)
    else:
        raise KeyError(("Table name must be one of: {}".format(list(sorted(available_table_names)))))



# Copyright (c) 2019-2021 Thomas Kramer.
# SPDX-FileCopyrightText: 2022 Thomas Kramer
#
# SPDX-License-Identifier: GPL-3.0-or-later

from .types import *
from .tokenized import *
from typing import *

class LibertyParserError(ParserError):
    pass

class ParseIntError(LibertyParserError):
    pass

class ParseFloatError(LibertyParserError):
    pass

class ExceptionWithLineNum(LibertyParserError):

    def __init__(self, e, line_num, char_num):
        self.e = e
        self.line_num = line_num
        self.char_num = char_num

    def __str__(self):
        return f"Error on line {self.line_num}, position {self.char_num}: {self.e.__repr__()}"

class _LibertyBuilder:
    """
    Construct liberty data structures while the parser encounters them.
    """

    def __init__(self):
        self._finished_groups: List[Group] = []
        self._group_stack: List[Group] = []
        self._cell_filter_fn: Callable[str, bool] = lambda name: True
        self._skip: int = 0

    def open_group(self, name: str, args: List):
        if self.skip_group():
            self._skip += 1
            return
        if len(self._group_stack) == 1:
            # library level
            if name == "cell" and len(args) > 0:
                if not self._cell_filter_fn(args[0]):
                    self._skip += 1
        if self.skip_group():
            return
        self._group_stack.append(Group(name, args))

    def push_group_item(self, item: Union[Group, Attribute, Define]):
        if self.skip_group():
            return
        group = self._group_stack[-1]
        if isinstance(item, Group):
            group.groups.append(item)
        elif isinstance(item, Attribute):
            group.attributes.append(item)
        elif isinstance(item, Define):
            group.defines.append(item)
        else:
            assert False, "unexpected type"

    def skip_group(self) -> bool:
        """
        Instruct the parser to skip through content of group without creating data structures.
        """
        return self._skip > 0

    def close_group(self):
        if self.skip_group():
            self._skip -= 1
            return
        g = self._group_stack.pop()
        if self._group_stack:
            self._group_stack[-1].groups.append(g)
        else:
            self._finished_groups.append(g)
    
class LibertyParser:

    def __init__(self):
        self.strict: bool = False
        """
        Be more strict and don't allow certain format variants which are not in line with the liberty reference manual.
        """

        self._cell_filter_fn = lambda name: True

    def set_cell_name_filter(self, take_cell_group_fn: Callable[str, bool]):
        """
        Decide wether to use or skip a cell group based on the cell name.

        Example: `parser.set_cell_name_filter(lambda name: name == 'INVX1')`
        """
        assert isinstance(take_cell_group_fn, Callable)
        self._cell_filter_fn = take_cell_group_fn

    def parse_liberty(self, data: str) -> Group:
        """
        Parse a string containing data of a liberty file.
        The liberty string must contain exactly one top library. If more than one top
        should be supported then `parse_multi_liberty()` should be used instead.

        :param data: Raw liberty string.
        :return: `Group` object of library.
        """
        top_groups = self.parse_multi_liberty(data)

        if len(top_groups) == 1:
            return top_groups[0]
        else:
            raise LibertyParserError("Liberty does not contain exactly one top group. Use `parse_multi_liberty()` instead.")

    def parse_multi_liberty(self, data: str) -> List[Group]:
        """
        Parse a string containing data of a liberty file.
        The liberty file may contain many top-level libraries.
        :param data: Raw liberty string.
        :return: List of `Group` objects.
        """
        library = self.read_liberty_chars(iter(data))
        return library

    def read_liberty_chars(self, chars: Iterable) -> List[Group]:
        """
        Parse liberty libraries from an iterator over characters.
        """
        assert isinstance(chars, Iterable)
    
        class CountLines:
            def __init__(self, iter):
                self.iter = iter
                self.line_num = 0
                self.char_num = 0 # Position on the line.
    
            def __iter__(self):
                return self
            
            def __next__(self):
                c = next(self.iter)
                self.char_num += 1
                if c == "\n":
                    self.line_num += 1
                    self.char_num = 0
                return c
    
        counted = CountLines(chars)
        
        try:
            result = _read_liberty_impl(counted, strict=self.strict, cell_filter_fn=self._cell_filter_fn)
        except Exception as e:
            raise ExceptionWithLineNum(e, counted.line_num, counted.char_num)
        
        return result



def parse_liberty(data: str) -> Group:
    """
    Parse a string containing data of a liberty file.
    The liberty string must contain exactly one top library. If more than one top
    should be supported then `parse_multi_liberty()` should be used instead.

    :param data: Raw liberty string.
    :return: `Group` object of library.
    """
    p = LibertyParser()
    return p.parse_liberty(data)


def parse_multi_liberty(data: str) -> List[Group]:
    """
    Parse a string containing data of a liberty file.
    The liberty file may contain many top-level libraries.
    :param data: Raw liberty string.
    :return: List of `Group` objects.
    """
    p = LibertyParser()
    return p.parse_multi_liberty(data)

def read_liberty_chars(chars: Iterable) -> List[Group]:
    """
    Parse liberty libraries from an iterator over characters.
    """
    p = LibertyParser()
    return p.read_liberty_chars(chars)


def _read_liberty_impl(
        chars: Iterable, 
        strict: bool = False, 
        cell_filter_fn: Callable[str, bool] = lambda name: True
    ) -> List[Group]:
    assert isinstance(chars, Iterable)
    tk = tokenize(chars, LibertyLexer())
    tk.advance()

    groups = []
    
    while True:
        builder = _LibertyBuilder()
        builder._cell_filter_fn = cell_filter_fn
        __read_group_item(builder, tk, strict=strict)
        assert len(builder._finished_groups) == 1
        item = builder._finished_groups[0]
        if not isinstance(item, Group):
            raise LibertyParserError("library must start with a group but found:", type(item))
        groups.append(item)

        if tk.current_token_ref() is None:
            # End of file.
            break

    return groups

def __read_group_item(builder: _LibertyBuilder, tk: Tokenized, strict: bool = False):
    assert isinstance(tk, Tokenized)
    assert isinstance(builder, _LibertyBuilder)

    name = tk.take_str()

    if tk.peeking_test_str("("):
        args = []
        # Allow ':' in tokens.
        tk.lexer.disable_terminal_char(':')
        try:
            tk.advance()
            # Group or complex attribute.
            while not tk.test_str(")"):
                args.append(__read_value(tk))
                if not tk.peeking_test_str(")"):
                    tk.expect_str(",")
        finally:
            tk.lexer.enable_terminal_char(':')

        if tk.test_str("{"):
            # It's a group.

            builder.open_group(name, args)

            while not tk.test_str("}"):
                # Recursively read group items.
                __read_group_item(builder, tk, strict=strict)

            builder.close_group()

            if not strict:
                tk.test_str(";")
        else:
            # It's a complex attribute or define statement.
            tk.test_str(";") # Consume optional trailing semicolon.
            if builder.skip_group():
                return

            if name == "define" and len(args) == 3:
                # Define statement

                # Values must be names or quoted names.

                strings = []
                for a in args:
                    if isinstance(a, EscapedString):
                        s = a.value
                    else:
                        s = str(a)
                    strings.append(s)

                attribute_name, group_name, attr_type = strings
                
                builder.push_group_item(
                    Define(attribute_name, group_name, attr_type)
                )
            else:
                # It's a complex attribute
                builder.push_group_item(
                    Attribute(name, args)
                )
    elif tk.test_str(":"):
        # Simple attribute.
        value = __read_value(tk)
        
        is_expression = value in ["(", "-", "!"] or tk.current_token_str() in ["*", "+", "-"]
        if is_expression:
            # Read expression. Something like `VDD * 0.5 + 0.1`.
            expr = [str(value)]

            while not tk.test_str(";"):
                expr.append(tk.take_str())

            builder.push_group_item(
                Attribute(name, ArithExpression(" ".join(expr)))
            )
        else:
            tk.test_str(";") # Read optional semicolon
            builder.push_group_item(
                Attribute(name, value)
            )
    else:
        raise UnexpectedToken("'(' | ':'", tk.current_token_str())

def __read_value(tk: Tokenized):
    assert isinstance(tk, Tokenized)

    s = tk.current_token_str()
    if not s:
        raise UnexpectedEndOfFile()
    
    first_char = s[0]
    last_char = s[-1]

    if (first_char.isnumeric() or first_char == "-") and last_char.isnumeric():
        # Read a number
        is_int = all(c.isnumeric() for c in s[1:])
        s = tk.take_str()
        if is_int:
            return __read_int(s)
        else:
            try:
                return __read_float(s)
            except ParseFloatError:
                return s
    if __is_number_with_unit(s):
        return __read_number_with_unit(tk.take_str())
    elif first_char == '"':
        # Quoted string.
        # Strip away the quotes.
        without_quotes = s[1:-1]
        tk.advance()
        return EscapedString(without_quotes)
    else:
        name = tk.take_str()

        if tk.test_str("["):
            buspins = tk.take_str()
            tk.expect_str("]")
            splitted = buspins.split(":")

            if len(splitted) == 1:
                return NameBitSelection(name, int(splitted[0]))
            if len(splitted) == 2:
                return NameBitSelection(name, int(splitted[0]), int(splitted[1]))
            else:
                raise LibertyParserError("Invalid bus pins: {}".format(splitted))

        else:
            return name

def __read_int(s: str):
    try:
        return int(s)
    except ValueError as e:
        raise ParseIntError(e)
    
def __read_float(s: str):
    try:
        return float(s)
    except ValueError as e:
        raise ParseFloatError(e)

def __read_number_with_unit(s: str):
    try:
        unit_len = 0
        for c in s[::-1]:
            if c.isalpha():
                unit_len += 1
            else:
                break

        num = s[:-unit_len]
        unit = s[-unit_len:]

        num = float(num)

        return WithUnit(num, unit)
    except ValueError as e:
        return s

def __is_number_with_unit(s: str):
    if len(s) < 1:
        return False
    first = s[0]
    if first.isnumeric() or first == "-":
        return s[-1].isalpha()
    else:
        False

def test_read_liberty():
    
    data = r"""
    /*
    Author: somebody
    */
library (myLib) {
  time_unit : 1ns;
  simpleattr_int : 1;
  simpleattr_neg_int : -1;
  simpleattr_float : -1.12e-3;
    simple_attribute1 : value1;
    simple_attribute2 : value2;
    simple_attribute2 : value3;
    complex_attribute1 (value1, "value 2");

    // Single line comment // does not end here / also not here

    /* comment with ** * characters */

    cell (invx1) {
        simple_attribute1 : value;
    }
}
"""

    result = read_liberty_chars(iter(data))
    print(result)
    assert isinstance(result[0], Group)

def test_parse_liberty_simple():
    data = r"""
library(test) { 
  time_unit : 1ns;
  string : "asdf";
  mygroup(a, b) {}
  empty() {}
  somegroup(a, b, c) {
    nested_group(d, e) {
        simpleattr_float : 1.2;
    }
  }
  simpleattr_int : 1;
  complexattr(a, b);
  define(myNewAttr, validinthisgroup, float);
  pin(A[25]) {}
  pin(B[32:0]) {}
  pin(C[0:0]) {}
}
"""
    library = parse_liberty(data)
    assert isinstance(library, Group)

    # Check attribute values.
    assert library.get_attribute('simpleattr_int') == 1
    assert library.get_attribute('complexattr') == ['a', 'b']

    # Format, parse, format and check that the result stays the same.
    str1 = str(library)
    library2 = parse_liberty(str1)
    str2 = str(library2)
    assert (str1 == str2)
    
def test_parse_liberty_no_space_before_colon():
    data = r"""
library(test) { 
  attr_name: value;
  }
"""
    library = parse_liberty(data)
    assert isinstance(library, Group)

    # Check attribute values.
    assert library.get_attribute('attr_name') == "value"


def test_parse_liberty_with_unit():
    data = r"""
library(test) { 
  time_unit : 1ns ;
}
"""
    library = parse_liberty(data)
    assert isinstance(library, Group)

    # Check values with unit.
    assert isinstance(library.get_attribute('time_unit'), WithUnit)
    assert library.get_attribute('time_unit').value == 1
    assert library.get_attribute('time_unit').unit == 'ns'

    # Format, parse, format and check that the result stays the same.
    str1 = str(library)
    library2 = parse_liberty(str1)
    str2 = str(library2)
    assert (str1 == str2)


def test_parse_liberty_with_multline():
    data = r"""
table(table_name2){ 
    str : "asd\
    f";
    index_1("1, 2, 3, 4, 5, 6, 7, 8"); 
    value("0001, 0002, 0003, 0004, \
    0005, 0006, 0007, 0008");
}
"""
    library = parse_liberty(data)
    assert isinstance(library, Group)

    str1 = str(library)
    library2 = parse_liberty(str1)
    str2 = str(library2)
    assert (str1 == str2)


def test_parse_liberty_statetable_multiline():
    # From https://codeberg.org/tok/liberty-parser/issues/6
    data = r"""
statetable ("CK E SE","IQ") {
	     table : "L L L : - : L ,\
	              L L H : - : H ,\
	              L H L : - : H ,\
	              L H H : - : H ,\
	              H - - : - : N " ;
	}
"""

    library = parse_liberty(data)
    assert isinstance(library, Group)

    str1 = str(library)
    library2 = parse_liberty(str1)
    str2 = str(library2)
    assert (str1 == str2)


def test_parse_liberty_with_define():
    data = r"""
group(test){ 
    define (a, b, c);
    define (x, y, z);
}
"""
    library = parse_liberty(data)
    assert isinstance(library, Group)
    assert isinstance(library.defines[0], Define)
    assert isinstance(library.defines[1], Define)

    str1 = str(library)
    library2 = parse_liberty(str1)
    str2 = str(library2)
    assert (str1 == str2)


def test_parse_liberty_multi_complex_attributes():
    data = r"""
group(test){ 
    define_group(g1, x);
    define_group(g2, z);
    voltage_map(VDD, 1.0);
    voltage_map(VSS, 0.0);
}
"""
    library = parse_liberty(data)
    assert isinstance(library, Group)

    # Check if `voltage_map` is parsed as expected.
    assert library.get_attributes('voltage_map')[0] == ['VDD', 1.0]
    assert library.get_attributes('voltage_map')[1] == ['VSS', 0.0]

    str1 = str(library)
    library2 = parse_liberty(str1)
    assert len(library.attributes) == 4
    str2 = str(library2)
    assert (str1 == str2)


def test_parse_liberty_freepdk():
    import os.path
    lib_file = os.path.join(os.path.dirname(__file__), '../../test_data/gscl45nm.lib')

    data = open(lib_file).read()

    library = parse_liberty(data)
    assert isinstance(library, Group)

    library_str = str(library)
    open('/tmp/lib.lib', 'w').write(library_str)
    library2 = parse_liberty(library_str)
    assert isinstance(library2, Group)
    library_str2 = str(library2)
    assert (library_str == library_str2)

    cells = library.get_groups('cell')

    invx1 = library.get_group('cell', 'XOR2X1')
    assert invx1 is not None

    pin_y = invx1.get_group('pin', 'Y')
    timings_y = pin_y.get_groups('timing')
    timing_y_a = [g for g in timings_y if g['related_pin'] == 'A'][0]
    assert timing_y_a['related_pin'] == 'A'

    array = timing_y_a.get_group('cell_rise').get_array('values')
    assert array.shape == (6, 6)

def test_parse_and_filter_liberty_freepdk():
    import os.path
    lib_file = os.path.join(os.path.dirname(__file__), '../../test_data/gscl45nm.lib')

    data = open(lib_file).read()

    parser = LibertyParser()
    parser.set_cell_name_filter(lambda name: name == 'INVX1')
    filtered_library = parser.parse_liberty(data)
    assert isinstance(filtered_library, Group)
    assert len([g for g in filtered_library.groups if g.group_name == 'cell']) == 1
    
def test_parse_liberty_openram():
    import os.path
    lib_file = os.path.join(os.path.dirname(__file__), '../../test_data/openram_sram_16x8_FF.lib')

    data = open(lib_file).read()

    library = parse_liberty(data)
    assert isinstance(library, Group)

def test_wire_load_model():
    """
    Test that multiple attributes with the same name don't overwrite eachother.
    See: https://codeberg.org/tok/liberty-parser/issues/7
    """

    data = r"""
    wire_load("1K_hvratio_1_4") {
        capacitance : 1.774000e-01;
        resistance : 3.571429e-03;
        slope : 5.000000;
        fanout_length( 1, 1.3207 );
        fanout_length( 2, 2.9813 );
        fanout_length( 3, 5.1135 );
        fanout_length( 4, 7.6639 );
        fanout_length( 5, 10.0334 );
        fanout_length( 6, 12.2296 );
        fanout_length( 8, 19.3185 );
    }
"""
    wire_load = parse_liberty(data)
    fanout_lengths = wire_load.get_attributes("fanout_length")
    assert isinstance(fanout_lengths, list)
    assert len(fanout_lengths) == 7
    expected_fanoutlength = [
        [1, 1.3207],
        [2, 2.9813],
        [3, 5.1135],
        [4, 7.6639],
        [5, 10.0334],
        [6, 12.2296],
        [8, 19.3185],
    ]
    assert fanout_lengths == expected_fanoutlength


def test_argument_with_dot():
    """
    Parse names with dots like `a.b`.
    """
    # Issue #10
    data = r"""
operating_conditions(ff28_1.05V_0.00V_0.00V_0.00V_125C_7y50kR){
}    
"""
    group = parse_liberty(data)

    assert group.args == ["ff28_1.05V_0.00V_0.00V_0.00V_125C_7y50kR"]


def test_complex_attribute_without_semicolon():
    """
    Parse complex attributes without trailing `;`.
    """
    # Issue #10
    data = r"""
library(){
    cplxAttr1(1)
    cplxAttr2(1, 2)
    cplxAttr3(3);
    cplxAttr4(4)
}
"""
    group = parse_liberty(data)

    assert len(group.attributes) == 4


def test_simple_attribute_without_semicolon():
    """
    Parse simple attributes without trailing `;`.
    """
    # Issue #10
    data = r"""
library(){
    simpleAttr1 : 1ps
    simpleAttr2 : 2;
    simpleAttr3 : 3
}
"""
    group = parse_liberty(data)

    assert len(group.attributes) == 3


def test_multi_top_level_libraries():
    """
    Parse files with more than one top-level library.
    """
    # Issue #10
    data = r"""
library(lib1){
}
library(lib2){
}
"""
    tops = parse_multi_liberty(data)
    assert isinstance(tops, list)
    assert len(tops) == 2


def test_define():
    # Issue #10
    data = r"""
    library(){
        define ("a", "b", "c");
        define (d, "e", f);
        define (g, h, i)
    }
    """
    group = parse_liberty(data)
    assert isinstance(group, Group)
    assert len(group.defines) == 3

    assert group.defines[0].attribute_name == "a"
    assert group.defines[0].group_name == "b"
    assert group.defines[0].attribute_type == "c"
    assert group.defines[1].attribute_name == "d"
    assert group.defines[1].group_name == "e"
    assert group.defines[1].attribute_type == "f"
    assert group.defines[2].attribute_name == "g"
    assert group.defines[2].group_name == "h"
    assert group.defines[2].attribute_type == "i"


def test_arithmetic_expressions():
    # Issue 10

    data = r"""
    input_voltage(cmos) {
        vil : 0.5 * VDD ;
        vih : 0.7 * VDD ;
        vimin : -0.5 ;
        vimax : VDD * 1.1 + 0.5 ;
    }
"""
    group = parse_liberty(data)
    assert isinstance(group, Group)
    assert len(group.attributes) == 4

    for attr in group.attributes:
        expr_str = attr.value
        if not isinstance(attr.value, float):
            assert isinstance(expr_str, ArithExpression)

            expr = expr_str.to_sympy_expression()
            print(expr)

    assert group.attributes[3].value.to_sympy_expression() == sympy.parse_expr("VDD * 1.1 + 0.5")

def test_single_letter_units():
    # Issue 11

    data = r"""
    test() {
        int_value : 1V ; 
        float_value : 2.5e-1A ;
    }
"""

    group = parse_liberty(data)
    assert isinstance(group, Group)
    assert len(group.attributes) == 2
    assert group.attributes[0].value == WithUnit(1, "V")
    assert group.attributes[1].value == WithUnit(0.25, "A")

def test_units_starting_with_E():
    # Issue 11

    data = r"""
    test() {
        int_value : 1eV ; 
        float_value : 2.5e-1EV ;
    }
"""

    group = parse_liberty(data)
    assert isinstance(group, Group)
    assert len(group.attributes) == 2
    assert group.attributes[0].value == WithUnit(1, "eV")
    assert group.attributes[1].value == WithUnit(0.25, "EV")

def test_group_without_argument():

    data = r"""
    my_group() {
        some_attribute : "123";
    }
    """
    
    group = parse_liberty(data)
    assert isinstance(group, Group)

    print(group.args)

    assert len(group.args) == 0

def test_group_arguments_with_colon():
    # Issue 15

    data = r"""
    group (name_with:colon) {
        my_attribute : true;

        simple_attribute : 1;

        simple_attribute_2 : 2;
    }    
"""
    group = parse_liberty(data)
    assert isinstance(group, Group)

    assert group.args[0] == "name_with:colon"

def test_multiline_with_backslash():
    
    data = r"""
    group() {
        array ( \
            "0, 0, 0", \
            "0, 0, 0" \
        );
    }
    """

    group = parse_liberty(data)
    assert isinstance(group, Group)

def test_bus_pins():
    data = r"""
        pin(A[1:3]) {}
    """
    group = parse_liberty(data)
    assert isinstance(group, Group)

    assert isinstance(group.args[0], NameBitSelection)
    assert group.args[0].sel1 == 1
    assert group.args[0].sel2 == 3
    

def test_invalid_bus_pins():

    data = r"""
        library () {

        cell (x) {
            pin(A[1:2:3]) {}
        }
    }
    """

    error = None
    try:
        group = parse_liberty(data)
    except Exception as e:
        error = e

    assert error is not None

def test_cell_filter():

    def filter(name: str) -> bool:
        return name in ["INVX1"]
    
    data = r"""
        library () {

        cell (skipme1) {
            value: 1;
            subgroup() {
                subsubgroup() {
                    value: 2;
                }
            }
        }

        cell (INVX1) {}

        cell (skipme2) {}
    }
    """

    parser = LibertyParser()
    parser.set_cell_name_filter(filter)
    group = parser.parse_liberty(data)
    assert len(group.groups) == 1
    
def test_cell_filter_takes_none():
    
    data = r"""
        library () {

        cell (INVX1) {}
    }
    """

    parser = LibertyParser()
    parser.set_cell_name_filter(lambda name: False) # Drop all cells
    group = parser.parse_liberty(data)
    assert len(group.groups) == 0




# SPDX-FileCopyrightText: 2024 Thomas Kramer
# SPDX-License-Identifier: GPL-3.0-or-later

from typing import Iterable
from .lexer import *

def tokenize(iter, lexer):
    return Tokenized(PeekingIterator(iter), lexer)

class ParserError(Exception):
    pass

class UnexpectedToken(ParserError):

    def __init__(self, expected, actual):
        self.expected = expected
        self.actual = actual
        
class InvalidLiteral(ParserError):

    def __init__(self, literal):
        self.literal = literal
        
class UnexpectedEndOfFile(ParserError):
    pass

class Tokenized:

    def __init__(self, iter: PeekingIterator, lexer):
        assert isinstance(iter, PeekingIterator)
        self.iter = iter
        self.lexer = lexer
        self.has_current = False
        self.current_token = []

    def advance(self):
        """
        Advance to the next token.
        """
        buffer = self.current_token
        buffer.clear()
        
        self.lexer.consume_next_token(self.iter, lambda c: buffer.append(c))

        if buffer:
            self.current_token = buffer
            self.has_current = True
        else:
            self.has_current = False

    def current_token_ref(self):
        if self.has_current:
            return self.current_token
        else:
            return None

    def current_token_str(self):
        if self.has_current:
            return "".join(self.current_token)
        else:
            return None

    def current_token_copy(self):
        if self.has_current:
            self.current_token.copy()
        else:
            return None

    def take(self):
        """
        Take and consume token.
        """
        t = self.current_token
        self.current_token = []
        self.advance()
        if t:
            return t
        else:
            return None

    def take_str(self):
        t = self.take()
        if t is not None:
            return "".join(t)
        else:
            return None
        
    def expect(self, expected):
        """
        Test if the current token has the `expected` value.
        If not, raise an `UnexpectedToken` exception.
        """
        assert isinstance(expected, list)
        
        if self.current_token:
            if self.current_token == expected:
                self.advance()
            else:
                raise UnexpectedToken(expected, self.current_token)
        else:
            raise UnexpectedEndOfFile()

    def expect_str(self, expected: str):
        assert isinstance(expected, str)
        return self.expect(list(expected))        

    def peeking_test(self, expected):
        """
        Test if the current token is equal to `expected`.
        Does not consume the token.
        """
        assert isinstance(expected, list)
        
        if self.current_token:
            return self.current_token == expected
        else:
            return False
        
    def peeking_test_str(self, expected):
        return self.peeking_test(list(expected))

    def test(self, expected):
        """
        Test if the current token is equal to `expected`.
        If it matches, consume it.
        """
        assert isinstance(expected, list)

        matches = self.peeking_test(expected)
        if matches:
            self.advance()
        return matches

    def test_str(self, expected):
        assert isinstance(expected, str)
        return self.test(list(expected))

    def skip_until(self, stop_pattern):
        """
        Consume tokens until and including the `stop_pattern`.
        """
        while not self.test(stop_pattern):
            self.advance()
            
        

def test_tokenized():

    class SimpleLexer:

        def consume_next_token(self, iter, output_fn):
            whitespace =  [" ", "\t", "\n", "\r"]

            try:
                while True:
                    # Consume all whitespace.
                    if iter.peek() in whitespace:
                        next(iter)
                    else:
                        break
            except StopIteration:
                pass
            
            # Consume until whitespace.
            for c in iter:
                if c in whitespace:
                    return
                output_fn(c)

    data = "here \n are \t some words  "
    tk = tokenize(iter(data), SimpleLexer())

    tk.advance()
    assert tk.peeking_test_str("here")

    # Test exception:
    did_raise_exception = False
    try:
        tk.expect_str("expected")
    except UnexpectedToken as e:
        did_raise_exception = True
        assert e.expected == list("expected")
        assert e.actual == list("here")
    assert did_raise_exception
    
    assert tk.test_str("here")
    assert tk.current_token_str() == "are"
    tk.expect_str("are")
    tk.expect_str("some")
    tk.expect_str("words")
